import os
import json

from typing import Literal
from uuid import uuid4
from zipfile import ZipFile
from xml.etree.ElementTree import fromstring, Element
from .gen_part import generate_part
from .gen_index import gen_index, NavPoint
from .i18n import I18N
from .template import Template


def generate_epub_file(
    from_dir_path: str,
    epub_file_path: str,
    lan: Literal["zh", "en"] = "zh") -> None:

  i18n = I18N(lan)
  template = Template()
  index_path = os.path.join(from_dir_path, "index.json")
  meta_path = os.path.join(from_dir_path, "meta.json")
  assets_path: str | None = os.path.join(from_dir_path, "assets")
  head_chapter_path = os.path.join(from_dir_path, "chapter.xml")

  toc_ncx: str | None = None
  nav_points: list[NavPoint] = []
  meta: dict = {}
  has_head_chapter: bool = os.path.exists(head_chapter_path)
  has_cover: bool = os.path.exists(os.path.join(from_dir_path, "cover.png"))

  if os.path.exists(meta_path):
    with open(meta_path, "r", encoding="utf-8") as f:
      meta = json.loads(f.read())

  if not os.path.exists(assets_path):
    assets_path = None

  if os.path.exists(index_path):
    toc_ncx, nav_points = gen_index(
      template=template,
      i18n=i18n,
      meta=meta,
      file_path=index_path,
      has_cover=has_cover,
      check_chapter_exits=lambda id: os.path.exists(
        os.path.join(from_dir_path, f"chapter_{id}.xml"),
      ),
    )
  with ZipFile(epub_file_path, "w") as file:
    _write_basic_files(
      file=file,
      template=template,
      i18n=i18n,
      meta=meta,
      nav_points=nav_points,
      assets_path=assets_path,
      has_cover=has_cover,
      has_head_chapter=has_head_chapter,
    )
    if toc_ncx is not None:
      file.writestr("OEBPS/toc.ncx", toc_ncx.encode("utf-8"))

    _write_assets(
      file=file,
      template=template,
      i18n=i18n,
      from_dir_path=from_dir_path,
      assets_path=assets_path,
      has_cover=has_cover,
    )
    _write_chapters(
      file=file,
      template=template,
      i18n=i18n,
      nav_points=nav_points,
      from_dir_path=from_dir_path,
      has_head_chapter=has_head_chapter,
      head_chapter_path=head_chapter_path,
    )

def _write_basic_files(
    file: ZipFile,
    template: Template,
    i18n: I18N,
    meta: dict,
    nav_points: list[NavPoint],
    assets_path: str | None,
    has_cover: bool,
    has_head_chapter: bool,
  ):

  file.writestr(
    zinfo_or_arcname="mimetype",
    data=template.render("mimetype").encode("utf-8"),
  )
  file.writestr(
    zinfo_or_arcname="META-INF/container.xml",
    data=template.render("container.xml").encode("utf-8"),
  )
  asset_files: list[str]
  if assets_path is None:
    asset_files = []
  else:
    asset_files = [
      f for f in os.listdir(assets_path)
      if not f.startswith(".")
    ]
  file.writestr(
    zinfo_or_arcname="OEBPS/content.opf",
    data=template.render(
      template="content.opf",
      meta=meta,
      i18n=i18n,
      ISBN=meta.get("ISBN", str(uuid4())),
      nav_points=nav_points,
      has_head_chapter=has_head_chapter,
      has_cover=has_cover,
      asset_files=asset_files,
    ).encode("utf-8"),
  )

def _write_assets(
    file: ZipFile,
    template: Template,
    i18n: I18N,
    from_dir_path: str,
    assets_path: str | None,
    has_cover: bool,
  ):
  file.writestr(
    zinfo_or_arcname="OEBPS/styles/style.css",
    data=template.render("style.css").encode("utf-8"),
  )
  if has_cover:
    file.writestr(
      zinfo_or_arcname="OEBPS/Text/cover.xhtml",
      data=template.render(
        template="cover.xhtml",
        i18n=i18n,
      ).encode("utf-8"),
    )

  if has_cover:
    file.write(
      filename=os.path.join(from_dir_path, "cover.png"),
      arcname="OEBPS/assets/cover.png",
    )

  if assets_path is not None:
    for file_name in sorted(os.listdir(assets_path)):
      file_path = os.path.join(assets_path, file_name)
      file.write(
        filename=file_path,
        arcname="OEBPS/assets/" + file_name,
      )

def _write_chapters(
    file: ZipFile,
    template: Template,
    i18n: I18N,
    nav_points: list[NavPoint],
    from_dir_path: str,
    has_head_chapter: bool,
    head_chapter_path: str,
):

  if has_head_chapter:
    chapter_xml = _read_xml(head_chapter_path)
    file.writestr(
      zinfo_or_arcname="OEBPS/Text/head.xhtml",
      data=generate_part(template, chapter_xml, i18n).encode("utf-8"),
    )
  for nav_point in nav_points:
    chapter_path = os.path.join(from_dir_path, f"chapter_{nav_point.index_id}.xml")
    if os.path.exists(chapter_path):
      chapter_xml = _read_xml(chapter_path)
      file.writestr(
        zinfo_or_arcname="OEBPS/Text/" + nav_point.file_name,
        data=generate_part(template, chapter_xml, i18n).encode("utf-8"),
      )

def _read_xml(path: str) -> Element:
  with open(path, "r", encoding="utf-8") as file:
    return fromstring(file.read())


from importlib.resources import files
from jinja2 import Environment, Template as JinjaTemplate
from ..template import create_env

class Template:
  def __init__(self):
    templates_path = files("pdf_craft").joinpath("data/templates")
    self._env: Environment = create_env(templates_path)
    self._templates: dict[str, JinjaTemplate] = {}

  def render(self, template: str, **params) -> str:
    template: JinjaTemplate = self._template(template)
    return template.render(**params)

  def _template(self, name: str) -> JinjaTemplate:
    template: JinjaTemplate = self._templates.get(name, None)
    if template is None:
      template = self._env.get_template(name)
      self._templates[name] = template
    return template


import os
import fitz

from html import escape
from hashlib import sha256
from typing import Generator
from PIL.Image import Image
from xml.etree.ElementTree import Element
from ..pdf import PDFPageExtractor, Block, Text, TextBlock, AssetBlock, TextKind, AssetKind
from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .asset_matcher import AssetMatcher, ASSET_TAGS


def extract_ocr_page_xmls(
    extractor: PDFPageExtractor,
    pdf_path: str,
    expected_page_indexes: set[int],
    cover_path: str,
    assets_dir_path: str,
    report_step: AnalysingStepReport | None,
    report_progress: AnalysingProgressReport | None,
  ) -> Generator[Element, None, None]:

  with fitz.open(pdf_path) as pdf:
    if report_step is not None:
      report_step(
        AnalysingStep.OCR,
        pdf.page_count - len(expected_page_indexes),
      )
    for i, blocks, image in extractor.extract_enumerated_blocks_and_image(
      pdf=pdf,
      page_indexes=(i for i in range(pdf.page_count) if i not in expected_page_indexes),
    ):
      if i == 0:
        image.save(cover_path)

      page_xml = _transform_page_xml(blocks)
      _bind_and_save_assets(
        root=page_xml,
        blocks=blocks,
        assets_dir_path=assets_dir_path,
      )
      yield i, page_xml

      if report_progress is not None:
        report_progress(i + 1)

def _transform_page_xml(blocks: list[Block]) -> Element:
  root = Element("page")
  for block in blocks:
    if isinstance(block, TextBlock):
      tag_name: str
      if block.kind == TextKind.TITLE:
        tag_name = "headline"
      elif block.kind == TextKind.PLAIN_TEXT:
        tag_name = "text"
      elif block.kind == TextKind.ABANDON:
        tag_name = "abandon"

      text_dom = Element(tag_name)
      if block.kind == TextKind.PLAIN_TEXT:
        text_dom.set("indent", "true" if block.has_paragraph_indentation else "false")
        text_dom.set("touch-end", "true" if block.last_line_touch_end else "false")

      _extends_line_doms(text_dom, block.texts)
      root.append(text_dom)

    elif isinstance(block, AssetBlock):
      tag_name: str
      if block.kind == AssetKind.FIGURE:
        tag_name = "figure"
      elif block.kind == AssetKind.TABLE:
        tag_name = "table"
      elif block.kind == AssetKind.FORMULA:
        tag_name = "formula"

      root.append(Element(tag_name))
      if len(block.texts) > 0:
        caption_dom = Element(f"{tag_name}-caption")
        _extends_line_doms(caption_dom, block.texts)
        root.append(caption_dom)

  return root

def _extends_line_doms(parent: Element, texts: list[Text]):
  for text in texts:
    content = text.content.replace("\n", " ")
    content = escape(content.strip())
    line_dom = Element("line")
    line_dom.set("confidence", "{:.2f}".format(text.rank))
    line_dom.text = content
    parent.append(line_dom)

def _bind_and_save_assets(root: Element, blocks: list[Block], assets_dir_path: str):
  asset_matcher = AssetMatcher()
  images: dict[str, Image] = {}
  for block in blocks:
    if isinstance(block, AssetBlock):
      hash = _block_image_hash(block)
      images[hash] = block.image
      asset_matcher.register_hash(block.kind, hash)
  asset_matcher.add_asset_hashes_for_xml(root)

  for hash in _handle_asset_tags(root):
    image: Image | None = images.get(hash, None)
    if image is not None:
      file_path = os.path.join(assets_dir_path, f"{hash}.png")
      if not os.path.exists(file_path):
        image.save(file_path, "PNG")

def _block_image_hash(block: AssetBlock) -> str:
  hash = sha256()
  hash.update(block.image.tobytes())
  return hash.hexdigest()

def _handle_asset_tags(parent: Element):
  pre_asset: Element | None = None
  asset_captions: list[Element] = []
  for child in parent:
    if child.tag not in ASSET_TAGS:
      if child.tag == "citation":
        _handle_asset_tags(child)
      if pre_asset is not None and \
         child.tag == f"{pre_asset.tag}-caption":
        for caption_child in child:
          pre_asset.append(caption_child)
        asset_captions.append(child)
      pre_asset = None
    if "hash" in child.attrib:
      pre_asset = child
      yield child.get("hash")
  for asset_caption in asset_captions:
    parent.remove(asset_caption)



from typing import Callable
from enum import auto, Enum


# func(completed_pages: int, all_pages: int) -> None
PDFPageExtractorProgressReport = Callable[[int, int], None]

class OCRLevel(Enum):
  Once = auto()
  OncePerLayout = auto()



from typing import Iterable, Generator
from xml.etree.ElementTree import fromstring, Element
from resource_segmentation import split, Resource, Incision

from ..llm import LLM
from .common import PageInfo
from .page_clipper import get_and_clip_pages, PageRef, PageXML
from .asset_matcher import AssetMatcher, ASSET_TAGS
from .chunk_file import ChunkFile
from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .utils import parse_page_indexes


def analyse_citations(
    llm: LLM,
    file: ChunkFile,
    pages: list[PageInfo],
    data_max_tokens: int,
    tail_rate: float,
    report_step: AnalysingStepReport | None,
    report_progress: AnalysingProgressReport | None,
  ) -> None:

  groups = file.filter_groups(split(
    max_segment_count=data_max_tokens,
    gap_rate=tail_rate,
    tail_rate=1.0,
    resources=_extract_citations(pages),
  ))
  if report_step is not None:
    groups = list(groups)
    report_step(AnalysingStep.EXTRACT_CITATION, len(groups))

  for i, (start_idx, end_idx, task_group) in enumerate(groups):
    page_xml_list = get_and_clip_pages(
      llm=llm,
      group=task_group,
      get_element=lambda i: _get_citation_with_file(pages, i),
    )
    raw_pages_root = Element("pages")
    for j, page_xml in enumerate(page_xml_list):
      element = page_xml.xml
      element.set("idx", str(j + 1))
      raw_pages_root.append(element)

    asset_matcher = AssetMatcher().register_raw_xml(raw_pages_root)
    response_xml = llm.request_xml("citation", raw_pages_root)
    chunk_xml = Element("chunk", {
      "start-idx": str(start_idx + 1),
      "end-idx": str(end_idx + 1),
    })
    asset_matcher.add_asset_hashes_for_xml(response_xml)

    for citation in _search_and_filter_and_split_citations(
      response_xml=response_xml,
      page_xml_list=page_xml_list,
    ):
      chunk_xml.append(citation)

    file.atomic_write_chunk(start_idx, end_idx, chunk_xml)

    if report_progress is not None:
      report_progress(i + 1)

def _extract_citations(pages: Iterable[PageInfo]) -> Generator[Resource[PageRef], None, None]:
  citations_matrix: list[list[Resource[PageRef]]] = []
  current_citations: list[Resource[PageRef]] = []

  for page in pages:
    citation = page.citation
    if citation is not None:
      current_citations.append(citation)
    elif len(current_citations) > 0:
      citations_matrix.append(current_citations)
      current_citations = []

  if len(current_citations) > 0:
    citations_matrix.append(current_citations)

  for citations in citations_matrix:
    citations[0].start_incision = Incision.IMPOSSIBLE
    citations[-1].end_incision = Incision.IMPOSSIBLE

  for citations in citations_matrix:
    yield from citations

def _get_citation_with_file(pages: list[PageInfo], index: int) -> Element:
  page = next((p for p in pages if p.page_index == index), None)
  assert page is not None
  assert page.citation is not None

  with page.file() as file:
    root: Element = fromstring(file.read())
    citation = root.find("citation")
    for child in citation:
      if child.tag not in ASSET_TAGS:
        child.attrib = {}
    return citation

def _search_and_filter_and_split_citations(response_xml: Element, page_xml_list: list[PageXML]):
  for citation in response_xml:
    page_indexes = parse_page_indexes(citation)
    page_xml = page_xml_list[page_indexes[0] - 1]

    if page_xml.is_gap:
      continue

    attributes = {
      **citation.attrib,
      "idx": ",".join([
        str(page_xml_list[p - 1].page_index + 1)
        for p in page_indexes
      ]),
    }
    # after testing, LLM will merge multiple citations together, which will result in multiple
    # labels for one citation. the code here handles this as a backup.
    splitted_citation: Element | None = None

    for child in citation:
      if child.tag == "label":
        if splitted_citation is not None:
          yield splitted_citation
        splitted_citation = Element("citation", {**attributes})

      if splitted_citation is not None:
        splitted_citation.append(child)

    if splitted_citation is not None:
      yield splitted_citation



from __future__ import annotations

import re
import os
import sys

from typing import Generator, Iterable
from xml.etree.ElementTree import fromstring, tostring, Element
from resource_segmentation import Resource, Group, Segment
from .common import PageRef


class ChunkFile:
  def __init__(self, output_dir_path: str):
    self._output_dir_path: str = output_dir_path
    self._files: list[tuple[int, int, str]] = list(self._search_chunk_file(output_dir_path))
    self._files.sort(key=lambda x: (x[0], x[1]))
    self._overlapped_files: dict[tuple[int, int], str] = {}
    self._min_page_index: int = sys.maxsize
    self._max_page_index: int = 0

  def __enter__(self) -> ChunkFile:
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    if exc_type is not None:
      return
    for file_name in self._overlapped_files.values():
      self._remove_file(file_name)
    for start, end, file_name in self._files:
      if start > self._max_page_index:
        self._remove_file(file_name)
      if end < self._min_page_index:
        self._remove_file(file_name)

  def filter_origin_files(self, origin_dir_path: str) -> Generator[tuple[int, int, Element], None, None]:
    for start, end, file_name in self._search_chunk_file(origin_dir_path):
      self._register_page_range(start, end)
      if self._overlap_files(start, end):
        file_path = os.path.join(origin_dir_path, file_name)
        with open(file_path, "r", encoding="utf-8") as file:
          yield start, end, fromstring(file.read())

  def _search_chunk_file(self, dir_path: str) -> Generator[tuple[int, int, str], None, None]:
    for file_name in os.listdir(dir_path):
      matches = re.match(r"^chunk_(\d+_\d+)\.xml$", file_name)
      if not matches:
        continue
      start, end = matches.group(1).split("_")
      yield int(start) - 1, int(end) - 1, file_name

  def filter_groups(self, groups: Iterable[Group[PageRef]]) -> Generator[tuple[int, int, Group], None, None]:
    for group in groups:
      start = min(t.payload.page_index for t in self._search_text_infos(group))
      end = max(t.payload.page_index for t in self._search_text_infos(group))
      self._register_page_range(start, end)

      if self._overlap_files(start, end):
        yield start, end, group

  def _search_text_infos(self, group: Group[PageRef]):
    for item in group.body:
      if isinstance(item, Segment):
        for text_info in item.resources:
          yield text_info
      elif isinstance(item, Resource):
        yield item

  def _register_page_range(self, start: int, end: int):
    self._min_page_index = min(self._min_page_index, start)
    self._max_page_index = max(self._max_page_index, end)

  # return False if matches origin chunk file
  def _overlap_files(self, start: int, end: int):
    overlap_files = list(self._search_overlap_files(start, end))
    overlap_files.sort(key=lambda x: (x[0], x[1]))

    if len(overlap_files) == 1:
      overlap_start, overlap_end, _ = overlap_files[0]
      if start == overlap_start and end == overlap_end:
        return False

    if len(overlap_files) > 0:
      self._overlapped_files[(start, end)] = overlap_files

    return True

  def _search_overlap_files(self, start: int, end: int):
    for file_start, file_end, file_name in self._files:
      if end < file_start:
        continue
      if start > file_end:
        continue
      yield file_start, file_end, file_name

  def atomic_write_chunk(self, start: int, end: int, chunk_xml: Element):
    overlap_files = self._overlapped_files.pop((start, end), None)
    if overlap_files is not None:
      for _, _, file_name in overlap_files:
        self._remove_file(file_name)

    file_name = f"chunk_{start + 1}_{end + 1}.xml"
    file_path = os.path.join(self._output_dir_path, file_name)
    try:
      with open(file_path, "w", encoding="utf-8") as file:
        content = tostring(chunk_xml, encoding="unicode")
        file.write(content)
        file.flush()
    except Exception as e:
      if os.path.exists(file_path):
        os.unlink(file_path)
      raise e

  def _remove_file(self, file_name: str):
    file_path = os.path.join(self._output_dir_path, file_name)
    if os.path.exists(file_path):
      os.remove(file_path)


import os
import re

from typing import Tuple, Callable
from jinja2 import select_autoescape, Environment, BaseLoader, TemplateNotFound

def create_env(dir_path: str) -> Environment:
  return Environment(
    loader=_DSLoader(dir_path),
    autoescape=select_autoescape(),
    trim_blocks=True,
    keep_trailing_newline=True,
  )

_LoaderResult = Tuple[str, str | None, Callable[[], bool] | None]

class _DSLoader(BaseLoader):
  def __init__(self, dir_path: str):
    super().__init__()
    self._dir_path: str = dir_path

  def get_source(self, _: Environment, template: str) -> _LoaderResult:
    template = self._norm_template(template)
    target_path = os.path.join(self._dir_path, template)
    target_path = os.path.abspath(target_path)

    if not os.path.exists(target_path):
      raise TemplateNotFound(f"cannot find {template}")

    return self._get_source_with_path(target_path)

  def _norm_template(self, template: str) -> str:
    if bool(re.match(r"^\.+/", template)):
      raise TemplateNotFound(f"invalid path {template}")

    template = re.sub(r"^/", "", template)
    template = re.sub(r"\.jinja$", "", template, flags=re.IGNORECASE)
    template = f"{template}.jinja"

    return template

  def _get_source_with_path(self, path: str) -> _LoaderResult:
    mtime = os.path.getmtime(path)
    with open(path, "r", encoding="utf-8") as f:
      source = f.read()

    def is_updated() -> bool:
      return mtime == os.path.getmtime(path)

    return source, path, is_updated


import os

from typing import Iterable
from xml.etree.ElementTree import fromstring, Element
from resource_segmentation import split, Resource, Incision

from ..llm import LLM
from .common import PageInfo, PageRef
from .chunk_file import ChunkFile
from .index import Index
from .page_clipper import get_and_clip_pages
from .asset_matcher import AssetMatcher
from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .utils import search_xml_and_indexes, parse_page_indexes


def analyse_main_texts(
    llm: LLM,
    file: ChunkFile,
    index: Index | None,
    pages: list[PageInfo],
    citations_dir_path: str,
    data_max_tokens: int, # TODO: not includes tokens of citations
    gap_rate: float,
    report_step: AnalysingStepReport | None,
    report_progress: AnalysingProgressReport | None,
  ):

  citations = _CitationLoader(citations_dir_path)
  groups = file.filter_groups(split(
    max_segment_count=data_max_tokens,
    gap_rate=gap_rate,
    tail_rate=0.5,
    resources=_extract_page_text_infos(pages),
  ))
  if report_step is not None:
    groups = list(groups)
    report_step(AnalysingStep.EXTRACT_MAIN_TEXT, len(groups))

  # TODO: 构建 summary class 用 spacy 循环删掉第一句话，以强行将它的范围限定在特定 tokens 数之内。
  #       LLM 的总结有时候会出问题，多次循环后 summary 就不变了。
  summary: str | None = None

  for i, (start_idx, end_idx, task_group)in enumerate(groups):
    page_xml_list = get_and_clip_pages(
      llm=llm,
      group=task_group,
      get_element=lambda i: _get_page_with_file(pages, i),
    )
    if index is not None:
      page_xml_list = [
        p for p in page_xml_list
        if not index.is_index_page_index(p.page_index)
      ]
      if len(page_xml_list) == 0:
        continue

    raw_pages_root = Element("pages")

    for j, page_xml in enumerate(page_xml_list):
      element = page_xml.xml
      element.set("idx", str(j + 1))
      citation = citations.load(page_xml.page_index)
      if citation is not None:
        element.append(citation)
      raw_pages_root.append(element)

    if summary is not None:
      summary_xml = Element("summary")
      summary_xml.text = summary
      raw_pages_root.append(summary_xml)

    asset_matcher = AssetMatcher().register_raw_xml(raw_pages_root)
    response_xml = llm.request_xml("main_text", raw_pages_root)
    chunk_xml = Element("chunk", {
      "start-idx": str(start_idx + 1),
      "end-idx": str(end_idx + 1),
    })
    asset_matcher.add_asset_hashes_for_xml(response_xml)
    abstract_xml = response_xml.find("abstract")
    assert abstract_xml is not None
    content_xml = Element("content")
    chunk_xml.append(abstract_xml)
    chunk_xml.append(content_xml)
    summary = abstract_xml.text

    for child in response_xml.find("content"):
      page_indexes = [
        j for j in parse_page_indexes(child)
        if 0 <= j < len(page_xml_list)
      ]
      if any(not page_xml_list[k].is_gap for k in page_indexes):
        attr_ids: list[str] = []
        for k in page_indexes:
          page_index = page_xml_list[k].page_index + 1
          attr_ids.append(str(page_index))
        child.set("idx", ",".join(attr_ids))
        content_xml.append(child)

    citation_xml = _collect_citations_and_reallocate_ids(raw_pages_root, chunk_xml)
    if citation_xml is not None:
      chunk_xml.append(citation_xml)

    file.atomic_write_chunk(start_idx, end_idx, chunk_xml)

    if report_progress is not None:
      report_progress(i + 1)

def _extract_page_text_infos(pages: list[PageInfo]) -> Iterable[Resource[PageRef]]:
  if len(pages) == 0:
    return

  previous: PageInfo | None = None

  for page in pages:
    if previous is not None and previous.page_index + 1 != page.page_index:
      previous.main.end_incision = Incision.IMPOSSIBLE
      previous = None
    if previous is None:
      page.main.start_incision = Incision.IMPOSSIBLE
    previous = page

  if previous is not None:
    previous.main.end_incision = Incision.IMPOSSIBLE

  for page in pages:
    yield page.main

def _get_page_with_file(pages: list[PageInfo], index: int) -> Element:
  page = next((p for p in pages if p.page_index == index), None)
  assert page is not None

  with page.file() as file:
    root: Element = fromstring(file.read())
    citation = root.find("citation")
    if citation is not None:
      root.remove(citation)
    return root

def _collect_citations_and_reallocate_ids(raw_xml: Element, chunk_xml: Element) -> Element | None:
  next_id: int = 1
  citations_map: dict[str, Element] = {}
  ids_map: dict[str, int] = {}
  keep_citations: list[Element] = []

  for page in raw_xml:
    if page.tag != "page":
      continue
    citations = page.find("citations")
    if citations is None:
      continue
    for citation in citations:
      id: str = citation.get("id")
      new_citation = Element("citation")
      citations_map[id] = new_citation
      for child in citation:
        new_citation.append(child)

  for ref in _search_refs(chunk_xml.find("content")):
    origin_id = ref.get("id")
    citation = citations_map.get(origin_id, None)
    if citation is None:
      continue

    id: str | None = ids_map.get(origin_id, None)
    if id is None:
      id = next_id
      next_id += 1
      ids_map[origin_id] = id

    citation.attrib = { "id": str(id) }
    ref.set("id", str(id))
    if citation not in keep_citations:
      keep_citations.append(citation)

  keep_citations.sort(key=lambda c: int(c.get("id")))
  if len(keep_citations) == 0:
    return None

  citations_xml = Element("citations")
  for citation in keep_citations:
    citations_xml.append(citation)
  return citations_xml

def _search_refs(parent: Element):
  for child in parent:
    if child.tag == "ref":
      yield child
    else:
      yield from _search_refs(child)

class _CitationLoader:
  def __init__(self, dir_path: str):
    self._next_citation_id: int = 1
    self._dir_path: str = dir_path
    self._index2file: dict[int, str] = {}

    for file_name, _, _ in search_xml_and_indexes("chunk", dir_path):
      file_path = os.path.join(dir_path, file_name)
      with open(file_path, "r", encoding="utf-8") as file:
        root = fromstring(file.read())
        for page_index in self._read_page_indexes(root):
          self._index2file[page_index] = file_name

  def _read_page_indexes(self, root: Element):
    for child in root:
      if child.tag == "citation":
        page_indexes = parse_page_indexes(child)
        if len(page_indexes) > 0:
          yield page_indexes[0]

  def load(self, page_index: int) -> Element | None:
    file_name = self._index2file.get(page_index, None)
    if file_name is None:
      return None

    root: Element
    file_path = os.path.join(self._dir_path, file_name)
    with open(file_path, "r", encoding="utf-8") as file:
      root = fromstring(file.read())

    citations = Element("citations")
    for child in root:
      if child.tag != "citation":
        continue
      page_indexes = parse_page_indexes(child)
      if len(page_indexes) == 0:
        continue
      if page_index != page_indexes[0]:
        continue
      child.attrib = { "id": str(self._next_citation_id) }
      citations.append(child)
      self._next_citation_id += 1

    return citations



from __future__ import annotations
import os
import io

from dataclasses import dataclass
from typing import Iterable, Generator
from xml.etree.ElementTree import fromstring, tostring, Element

from ..llm import LLM
from .asset_matcher import ASSET_TAGS
from .utils import read_files, normalize_xml_text, search_xml_children, parse_page_indexes


def serials(llm: LLM, chunks_path: str) -> Generator[Serial, None, None]:
  yield from _Deduplication(llm, chunks_path).for_serials()

@dataclass
class Serial:
  main_texts: list[Element]
  citations: Citations

@dataclass
class Citation:
  id: int
  label: str
  content: list[Element]
  str_text: str

class Citations:
  def __init__(self):
    self._refs: dict[int, tuple[int, Citation]] = {}
    self._max_id: int = 0

  def __iter__(self) -> Generator[Citation, None, None]:
    for _, citation in self._refs.values():
      yield citation

  def get(self, id: int) -> Citation | None:
    ref = self._refs.get(id, None)
    if ref is None:
      return None
    return ref[1]

  # deduplication: will return citation with different id if the citation is already in the list
  def ref(self, id: int, label: str, content: list[Element]) -> Citation:
    str_text = "".join(tostring(e, encoding="unicode") for e in content)
    for count, citation in self._refs.values():
      if citation.label != label or citation.str_text != str_text:
        continue
      self._refs[citation.id] = (count + 1, citation)
      return citation

    if id in self._refs:
      id = self._max_id + 1
    citation = Citation(id, label, content, str_text)
    self._refs[id] = (1, citation)
    self._max_id = max(self._max_id, id)

    return citation

  def unref(self, id: int) -> Citation:
    assert id in self._refs, f"Cannot find citation with id {id}"
    count, citation = self._refs[id]
    if count == 1:
      self._refs.pop(id, None)
    else:
      self._refs[id] = (count - 1, citation)
    return citation

@dataclass
class _Chunk:
  file_name: str
  start_idx: int
  end_idx: int
  index: int
  serial: Serial | None

class _Deduplication:
  def __init__(self, llm: LLM, chunks_path: str):
    self._llm: LLM = llm
    self._chunks_path: str = chunks_path
    self._chunks: list[_Chunk] = [
      _Chunk(
        file_name=file_name,
        start_idx=index1 - 1,
        end_idx=index2 - 1,
        index=-1,
        serial=None,
      )
      for file_name, _, index1, index2 in read_files(
        dir_path=chunks_path,
        enable_kinds=("chunk",),
      )
    ]
    self._chunks.sort(key=lambda chunk: chunk.start_idx)
    for i, chunk in enumerate(self._chunks):
      chunk.index = i

  def for_serials(self) -> Generator[Serial, None, None]:
    for index, chunk in enumerate(self._chunks):
      serial = self._load_serial_and_deduplicate(index, chunk)
      chunk.serial = None
      if serial is not None:
        yield self._clean_all_idx_attr_with_serial(serial)

  def _load_serial_and_deduplicate(self, index: int, chunk: _Chunk) -> Serial | None:
    serial = self._load_serial(chunk)
    latest_text = self._find_end_text(serial, False)

    if latest_text is not None:
      duplicated = list(self._find_duplicated_texts_from_serials(latest_text, index))
      if len(duplicated) > 0:
        latest_text_index = serial.main_texts.index(latest_text)
        duplicated.insert(0, (latest_text, serial))
        merged_text, citations = self._remove_and_merge_texts_from_serials(duplicated)
        serial.main_texts.insert(latest_text_index, merged_text)

        for id, ref in self._search_refs_in_text(merged_text):
          citation = citations[id]
          deduplicated_citation = serial.citations.ref(
            id=id,
            label=citation.label,
            content=citation.content,
          )
          if id != deduplicated_citation.id:
            ref.set("id", str(deduplicated_citation.id))

    if len(serial.main_texts) == 0:
      # cleared due to deduplication
      return None
    else:
      return serial

  def _find_duplicated_texts_from_serials(self, text: Element, index: int):
    ban_max_index = index # the processed index cannot be processed again
    search_indexes = self._chunk_indexes_with_text(text, index)

    while len(search_indexes) > 0:
      next_index = search_indexes.pop(0)
      serial = self._load_serial(self._chunks[next_index])
      first_text = self._find_end_text(serial, True)
      if first_text is None:
        # If the process breaks down, it means that LLM made an error in judgment
        # and the process must be interrupted (this will not happen under every thing is OK)
        break

      next_indexes = self._chunk_indexes_with_text(first_text, next_index)
      if index not in next_indexes:
        # This means that the index is not in the same order as the current one.
        # Something must have gone wrong. To be on the safe side, end this operation.
        break

      yield first_text, serial

      origin_indexes_count = len(search_indexes)
      for i in next_indexes:
        if i in search_indexes:
          continue
        if i <= ban_max_index:
          continue
        search_indexes.append(i)

      if origin_indexes_count != len(search_indexes):
        search_indexes.sort()

      index = next_index
      ban_max_index = max(index, ban_max_index)

  def _remove_and_merge_texts_from_serials(self, duplicated: list[tuple[Element, Serial]]):
    citation_matrix: list[dict[int, Citation]] = []
    for text, serial in duplicated:
      citations: dict[int, Citation] = {}
      citation_matrix.append(citations)
      for id, _ in self._search_refs_in_text(text):
        citation = serial.citations.unref(id)
        citations[id] = citation
      serial.main_texts.remove(text)

    index = self._try_to_choose_from_texts(e[0] for e in duplicated)
    text, _ = duplicated[index]
    citations = citation_matrix[index]

    return text, citations

  def _try_to_choose_from_texts(self, texts: Iterable[Element]) -> int:
    str_texts: list[str] = []
    for text in texts:
      buffer = io.StringIO()
      buffer.write(normalize_xml_text(text.text))
      for child in text:
        buffer.write("<ref/>")
        if child.tail is not None:
          buffer.write(normalize_xml_text(child.tail))
      str_texts.append(buffer.getvalue())

    no_sub_indexes: list[int] = []
    for i, str_text1 in enumerate(str_texts):
      not_sub = True
      for j, str_text2 in enumerate(str_texts):
        if i != j and str_text1 in str_text2:
          not_sub = False
          break
      if not_sub:
        no_sub_indexes.append(i)

    if len(no_sub_indexes) != 1:
      # TODO: use LLM to choose the best text index
      max_len: int = 0
      max_len_index: int = -1
      for i, str_text in enumerate(str_texts):
        text_len = len(str_text)
        if text_len > max_len:
          max_len = text_len
          max_len_index = i
      return max_len_index

    return no_sub_indexes[0]

  def _load_serial(self, chunk: _Chunk):
    if chunk.serial is not None:
      return chunk.serial

    file_path = os.path.join(self._chunks_path, chunk.file_name)
    with open(file_path, "r", encoding="utf-8") as file:
      chunk_xml = fromstring(file.read())

    content_xml = chunk_xml.find("content")
    main_texts: list[Element] = []
    for child in content_xml:
      main_texts.append(child)

    citations_xml = chunk_xml.find("citations")
    citations = Citations()

    if citations_xml is not None:
      for citation_xml in citations_xml:
        id = int(citation_xml.get("id"))
        label_xml: Element | None = None
        content: list[Element] = []
        for child in citation_xml:
          if child.tag == "label":
            label_xml = child
          else:
            content.append(child)

        assert label_xml is not None, "Citation must have a label"
        citation = citations.ref(
          id=id,
          label=label_xml.text,
          content=content,
        )
        if citation.id != id: # be deduplicated
          citation_xml.set("id", str(citation.id))

    serial = Serial(main_texts, citations)
    chunk.serial = serial

    return serial

  def _find_end_text(self, serial: Serial, is_begin_end: bool) -> Element | None:
    main_texts = serial.main_texts
    range_iter: Iterable[int]

    if is_begin_end:
      range_iter = range(len(main_texts))
    else:
      range_iter = range(len(main_texts) - 1, -1, -1)

    for i in range_iter:
      element = main_texts[i]
      if element.tag == "text":
        return element
      elif element.tag not in ASSET_TAGS:
        # it is normal to insert figures, tables, and formulas and split text
        return None

    return None

  def _search_refs_in_text(self, text: Element):
    for target, _ in search_xml_children(text):
      if target.tag == "ref":
        id = int(target.get("id"))
        yield id, target

  def _clean_all_idx_attr_with_serial(self, serial: Serial):
    for child in serial.main_texts:
      self._clean_all_idx_attr(child)
    for citation in serial.citations:
      for child in citation.content:
        self._clean_all_idx_attr(child)
    return serial

  def _chunk_indexes_with_text(self, text: Element, expected_chunk_index: int):
    chunk_indexes: list[int] = []
    for page_index in parse_page_indexes(text):
      for chunk_index, chunk in enumerate(self._chunks):
        if chunk_index != expected_chunk_index and \
           chunk_index not in chunk_indexes and \
           chunk.start_idx <= page_index <= chunk.end_idx:
          chunk_indexes.append(chunk_index)
    chunk_indexes.sort()
    return chunk_indexes

  def _clean_all_idx_attr(self, element: Element):
    for target, _ in search_xml_children(element):
      target.attrib.pop("idx", None)
    element.attrib.pop("idx", None)



import os
import re

from typing import Iterable, Generator
from xml.etree.ElementTree import fromstring, Element


def normalize_xml_text(xml_text: str) -> str:
  return re.sub(r"\s+", " ", xml_text).strip()

def read_xml_files(dir_path: str, enable_kinds: Iterable[str]) -> Generator[tuple[Element, str, str, int, int], None, None]:
  for file_name, kind, index1, index2 in read_files(dir_path, enable_kinds):
    file_path = os.path.join(dir_path, file_name)
    with open(file_path, "r", encoding="utf-8") as file:
      root = fromstring(file.read())
      yield root, file_name, kind, index1, index2

def search_xml_and_indexes(kind: str, dir_path: str) -> Generator[tuple[str, int, int], None, None]:
  for file_name in os.listdir(dir_path):
    matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
    if not matches:
      continue
    file_kind: str
    index1: str
    index2: str
    cells = re.sub(r"\..*$", "", file_name).split("_")
    if len(cells) == 3:
      file_kind, index1, index2 = cells
    else:
      file_kind, index1 = cells
      index2 = index1
    if kind != file_kind:
      continue
    yield file_name, int(index1) - 1, int(index2) - 1

def read_files(dir_path: str, enable_kinds: Iterable[str]) -> Generator[tuple[str, str, int, int], None, None]:
  for file_name in os.listdir(dir_path):
    matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
    if not matches:
      continue

    kind: str
    index1: str
    index2: str = "-1"
    cells = re.sub(r"\..*$", "", file_name).split("_")

    if len(cells) == 3:
      kind, index1, index2 = cells
    else:
      kind, index1 = cells

    if kind not in enable_kinds:
      continue

    yield file_name, kind, int(index1), int(index2)

def search_xml_children(parent: Element) -> Generator[tuple[Element, Element], None, None]:
  for child in parent:
    yield child, parent
    yield from search_xml_children(child)

def parse_page_indexes(element: Element) -> list[int]:
  idx = element.get("idx")
  if idx is None:
    return []
  page_indexes = [int(i) - 1 for i in idx.split(",")]
  page_indexes.sort()
  return page_indexes

def group_range(indexes: Iterable[int]) -> Generator[range, None, None]:
  indexes = list(indexes)
  indexes.sort()
  current_range: tuple[int, int] | None = None

  for index in indexes:
    if current_range is None:
      current_range = (index, index)
    else:
      start, end = current_range
      if end + 1 == index:
        current_range = (start, index)
      else:
        yield range(start, end + 1)
        current_range = (index, index)

  if current_range is not None:
    start, end = current_range
    yield range(start, end + 1)


from xml.etree.ElementTree import Element
from ..llm import LLM


def extract_meta(llm: LLM, page_xmls: list[Element]) -> dict:
  raw_pages_xml = Element("pages")
  for i, page_xml in enumerate(page_xmls):
    raw_pages_xml.append(page_xml)
    page_xml.set("idx", str(i + 1))

  response_json = llm.request_json("meta", raw_pages_xml)
  if not isinstance(response_json, dict):
    return None
  if len(response_json) == 0:
    return None

  return response_json


from typing import Generator
from xml.etree.ElementTree import Element
from ..pdf import AssetKind


ASSET_TAGS = ("figure", "table", "formula")

class AssetMatcher:
  def __init__(self):
    self._asset_hashes: dict[AssetKind, list[str]] = {}

  def register_raw_xml(self, root: Element) -> "AssetMatcher":
    for element in search_asset_tags(root):
      kind = self._tag_to_asset_kind(element.tag)
      hash = element.get("hash")
      if hash is not None:
        self.register_hash(kind, hash)
    return self

  def register_hash(self, kind: AssetKind, hash: str):
    hashes = self._asset_hashes.get(kind, None)
    if hashes is None:
      hashes = []
      self._asset_hashes[kind] = hashes
    hashes.append(hash)

  def add_asset_hashes_for_xml(self, root: Element):
    for element in search_asset_tags(root):
      kind = self._tag_to_asset_kind(element.tag)
      hashes = self._asset_hashes.get(kind, None)
      hash: str | None = None
      if hashes:
        hash = hashes.pop(0)
      if hash is not None:
        element.set("hash", hash)

  def _tag_to_asset_kind(self, tag_name: str) -> AssetKind:
    if tag_name == "figure":
      return AssetKind.FIGURE
    elif tag_name == "table":
      return AssetKind.TABLE
    elif tag_name == "formula":
      return AssetKind.FORMULA
    else:
      raise ValueError(f"Unknown tag name: {tag_name}")

def search_asset_tags(target: Element) -> Generator[Element, None, None]:
  for child in target:
    if child.tag in ASSET_TAGS:
      yield child
    else:
      yield from search_asset_tags(child)


from __future__ import annotations
from typing import Iterable, Callable
from dataclasses import dataclass
from xml.etree.ElementTree import Element
from ..llm import LLM
from .utils import group_range, normalize_xml_text


def analyse_index(llm: LLM, raw: Iterable[tuple[int, Element]]) -> dict | None:
  raw_index_pages: list[tuple[int, Element]] = []
  for i, raw_page_xml in raw:
    if raw_page_xml.tag == "index":
      raw_index_pages.append((i, raw_page_xml))

  if len(raw_index_pages) == 0:
    return None, None

  raw_page_xml = Element("index")
  for i, (_, raw_index_xml) in enumerate(raw_index_pages):
    raw_index_xml.set("page-index", str(i + 1))
    raw_page_xml.append(raw_index_xml)

  response_xml = llm.request_xml("index", raw_page_xml)
  index_json = _transform_llm_response_to_json(response_xml)
  ranges = group_range(i + 1 for i, _ in raw_index_pages)
  index_json["ranges"] = [[r[0], r[-1]] for r in ranges]

  return index_json, Index(index_json)

def _transform_llm_response_to_json(response_xml: Element) -> dict:
  prefaces: Element | None = None
  chapters: Element | None = None

  for child in response_xml:
    if child.tag == "prefaces":
      prefaces = child
    elif child.tag == "chapters":
      chapters = child

  if chapters is None:
    if prefaces is None:
      return None
    chapters = prefaces

  previous_id: int = 0
  def gen_id() -> int:
    nonlocal previous_id
    previous_id += 1
    return previous_id

  return {
    "prefaces": _parse_chapters(prefaces, gen_id),
    "chapters": _parse_chapters(chapters, gen_id),
  }

def _parse_chapters(parent: Iterable[Element] | None, gen_id: Callable[[], int]) -> list[Chapter]:
  if parent is None:
    return []

  chapters: list[dict] = []
  for chapter in parent:
    if chapter.tag != "chapter":
      continue

    headline = chapter.find("headline")
    children = chapter.find("children")
    if headline is None:
      continue

    id = gen_id()
    chapters.append({
      "id": str(id),
      "headline": normalize_xml_text(headline.text),
      "children": _parse_chapters(children, gen_id) if children is not None else [],
    })
  return chapters

@dataclass
class Chapter:
  id: int
  headline: str
  children: list[Chapter]

class Index:
  def __init__(self, json_data: dict):
    self._prefaces: list[Chapter] = [self._parse_chapter(c) for c in json_data["prefaces"]]
    self._chapters: list[Chapter] = [self._parse_chapter(c) for c in json_data["chapters"]]
    self._ranges: list[range] = [range(r[0], r[1] + 1) for r in json_data["ranges"]]
    assert len(self._ranges) > 0

  def _parse_chapter(self, data: dict) -> Chapter:
    return Chapter(
      id=int(data["id"]),
      headline=data["headline"],
      children=[self._parse_chapter(child) for child in data["children"]],
    )

  def is_index_page_index(self, page_index: int) -> bool:
    return any(page_index in range for range in self._ranges)

  def after_first_index_page(self, page_index: int) -> bool:
    return page_index > self._ranges[0][-1]

  @property
  def json(self) -> dict:
    if len(self._prefaces) == 0:
      return self._json_list(self._chapters)
    else:
      return {
        "prefaces": self._json_list(self._prefaces),
        "chapters": self._json_list(self._chapters),
      }

  def _json_list(self, chapters: list[Chapter]) -> list[dict]:
    return [
      {
        "id": str(chapter.id),
        "headline": chapter.headline,
        "children": self._json_list(chapter.children),
      }
      for chapter in chapters
    ]



import os
import fitz

from typing import Generator, Literal, Iterable, Sequence
from dataclasses import dataclass
from PIL.Image import frombytes, Image
from doc_page_extractor import plot, Layout, DocExtractor, ExtractedResult
from .section import Section
from .types import OCRLevel, PDFPageExtractorProgressReport


# section can be viewed up to 2 pages back
_MAX_VIEWED_PAGES: int = 2

@dataclass
class DocumentParams:
  pdf: str | fitz.Document
  page_indexes: Iterable[int] | None
  report_progress: PDFPageExtractorProgressReport | None

class DocumentExtractor:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      model_dir_path: str,
      ocr_level: OCRLevel,
      debug_dir_path: str | None = None,
    ):
    self._debug_dir_path: str | None = debug_dir_path
    self._doc_extractor = DocExtractor(
      device=device,
      model_dir_path=model_dir_path,
      order_by_layoutreader=False,
      ocr_for_each_layouts=(ocr_level == OCRLevel.OncePerLayout),
    )

  def extract(self, params: DocumentParams) -> Generator[tuple[int, ExtractedResult, list[Layout]], None, None]:
    for result, section in self._extract_results_and_sections(params):
      framework_layouts = section.framework()
      yield section.page_index, result, [
        layout for layout in result.layouts
        if layout not in framework_layouts
      ]

  def _extract_results_and_sections(self, params: DocumentParams):
    queue: list[tuple[ExtractedResult, Section]] = []

    for page_index, result in self._extract_page_result(params):
      section = Section(page_index, result.layouts)
      for i, (_, pre_section) in enumerate(queue):
        offset = len(queue) - i
        pre_section.link_next(section, offset)

      queue.append((result, section))
      if len(queue) > _MAX_VIEWED_PAGES:
        yield queue.pop(0)

    for result, section in queue:
      yield result, section

  def _extract_page_result(self, params: DocumentParams):
    if self._debug_dir_path is not None:
      os.makedirs(self._debug_dir_path, exist_ok=True)

    document: fitz.Document
    should_close = False
    report_progress = params.report_progress

    if isinstance(params.pdf, str):
      document = fitz.open(params.pdf)
      should_close = True
    else:
      document = params.pdf

    scan_indexes, enable_indexes = self._page_indexes_range(
      document=document,
      page_indexes=params.page_indexes,
    )
    try:
      for i, page_index in enumerate(scan_indexes):
        dpi = 300 # for scanned book pages
        page = document.load_page(page_index)
        image = self._page_screenshot_image(page, dpi)
        result = self._doc_extractor.extract(
          image=image,
          adjust_points=False,
        )
        if self._debug_dir_path is not None:
          self._generate_plot(image, page_index, result, self._debug_dir_path)

        if page_index in enable_indexes:
          yield page_index, result

        if report_progress is not None:
          report_progress(i + 1, len(scan_indexes))

    finally:
      if should_close:
        document.close()

  def _page_indexes_range(self, document: fitz.Document, page_indexes: Iterable[int] | None) -> tuple[Sequence[int], Sequence[int]]:
    pages_count = document.page_count
    if page_indexes is None:
      return range(pages_count), range(pages_count)

    enable_set: set[int] = set()
    scan_set: set[int] = set()

    for i in page_indexes:
      if 0 <= i < document.page_count:
        enable_set.add(i)
        for j in range(i - _MAX_VIEWED_PAGES, i + _MAX_VIEWED_PAGES + 1):
          if 0 <= j < document.page_count:
            scan_set.add(j)

    enable_list: list[int] = list(enable_set)
    scan_list: list[int] = list(scan_set)
    enable_list.sort()
    scan_list.sort()

    return enable_list, scan_list

  def _page_screenshot_image(self, page: fitz.Page, dpi: int):
    default_dpi = 72
    matrix = fitz.Matrix(dpi / default_dpi, dpi / default_dpi)
    pixmap = page.get_pixmap(matrix=matrix)
    return frombytes("RGB", (pixmap.width, pixmap.height), pixmap.samples)

  def _generate_plot(self, image: Image, index: int, result: ExtractedResult, plot_path: str):
    plot_image: Image
    if result.adjusted_image is None:
      plot_image = image.copy()
    else:
      plot_image = result.adjusted_image

    plot(plot_image, result.layouts)
    os.makedirs(plot_path, exist_ok=True)
    image_path = os.path.join(plot_path, f"plot_{index + 1}.png")
    plot_image.save(image_path)


from setuptools import setup, find_packages

setup(
  name="pdf-craft",
  version="0.0.18",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  url="https://github.com/oomol-lab/pdf-craft",
  description="PDF craft can convert PDF files into various other formats. This project will focus on processing PDF files of scanned books. The project has just started.",
  packages=find_packages(),
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  include_package_data=True,
  classifiers=[
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.10",
  ],
  package_data={
    "pdf_craft": ["data/**/*.jinja"],
  },
  install_requires=[
    "jinja2>=3.1.5,<4.0.0",
    "pyMuPDF>=1.25.3,<2.0.0",
    "alphabet-detector>=0.0.7,<1.0.0",
    "shapely>=2.0.6,<3.0.0",
    "tiktoken>=0.9.0,<1.0.0",
    "doc-page-extractor==0.0.9",
    "resource-segmentation==0.0.1",
    "langchain[openai]>=0.3.21,<0.4.0",
  ],
)


from xml.etree.ElementTree import tostring, Element
from .i18n import I18N
from .template import Template


def generate_part(template: Template, chapter_xml: Element, i18n: I18N) -> str:
  content_xml = chapter_xml.find("content")
  citations_xml = chapter_xml.find("citations")
  assert content_xml is not None
  return template.render(
    template="part.xhtml",
    i18n=i18n,
    content=list(_render_content(content_xml)),
    citations=list(_render_citations(citations_xml)),
  )

def _render_content(content_xml: Element):
  used_ref_ids: set[str] = set()
  for child in content_xml:
    for to_element in _create_main_text_element(child, used_ref_ids):
      yield tostring(to_element, encoding="unicode")

def _render_citations(citations_xml: Element | None):
  if citations_xml is None:
    return

  for citation in citations_xml:
    to_div = Element("div")
    to_div.attrib["class"] = "citation"
    id = citation.get("id", None)
    is_first_child = True
    citation_children = [c for c in citation if c.tag != "label"]

    if len(citation_children) == 1:
      citation_children[0].tag = "text"

    used_citation_ids: set[str] = set()

    for child in citation_children:
      for to_element in _create_main_text_element(child):
        ref_element = Element("a")
        ref_element.text = f"[{id}]"
        ref_element.attrib = {
          "href": f"#ref-{id}",
          "class": "citation",
        }
        if id not in used_citation_ids:
          used_citation_ids.add(id)
          ref_element.attrib = {
            "id": f"citation-{id}",
            **ref_element.attrib,
          }
        if is_first_child:
          is_first_child = False
          if to_element.tag == "p":
            ref_element.tail = to_element.text
            to_element.text = None
            to_element.append(ref_element)
          else:
            injected_element = Element("p")
            to_div.append(injected_element)
            injected_element.append(ref_element)

        to_div.append(to_element)

    yield tostring(to_div, encoding="unicode")

def _create_main_text_element(origin: Element, used_ref_ids: set[str] | None = None):
  html_tag: str | None = None
  if origin.tag == "text":
    html_tag = "p"
  elif origin.tag == "quote":
    html_tag = "p"
  elif origin.tag == "headline":
    html_tag = "h1"

  if html_tag is not None:
    element = Element(html_tag)
    _fill_text_and_citations(element, origin, used_ref_ids)
    if origin.tag == "quote":
      blockquote = Element("blockquote")
      blockquote.append(element)
      yield blockquote
    else:
      yield element
    return

  hash = origin.get("hash", None)
  if hash is None:
    return

  image = Element("img")
  image.set("src", f"../assets/{hash}.png")

  alt: str | None = None
  if origin.text != "":
    alt = origin.text
  if alt is None:
    image.set("alt", "image")
  else:
    image.set("alt", alt)

  wrapper_div = Element("div")
  wrapper_div.set("class", "alt-wrapper")
  wrapper_div.append(image)

  if alt is not None:
    alt_div = Element("div")
    alt_div.set("class", "alt")
    alt_div.text = alt
    wrapper_div.append(alt_div)

  yield wrapper_div

def _fill_text_and_citations(element: Element, origin: Element, used_ref_ids: set[str] | None):
  element.text = origin.text
  for child in origin:
    if child.tag != "ref":
      continue
    id = child.get("id")
    assert id is not None
    anchor = Element("a")
    anchor.attrib = {
      "id": f"ref-{id}",
      "href": f"#citation-{id}",
      "class": "super",
    }
    if used_ref_ids is not None:
      if id in used_ref_ids:
        anchor.attrib.pop("id", None)
      else:
        used_ref_ids.add(id)

    anchor.text = f"[{id}]"
    anchor.tail = child.tail
    element.append(anchor)